package miniprologfd.solver

// note the renaming that is being done by these imports
import scala.collection.mutable.{ Map => MMap }
import miniprologfd.interpreter.{ Value, NumValue, SymPlaceholder => Variable }
import miniprologfd.syntax.lowlevel.{ SymbolicROP => ROP, SymEQ => EQ,
  SymNE => NE, SymLT => LT, SymLE => LE, SymGT => GT, SymGE => GE }


// an interval of integers from 0 to Intervals.maxValue, inclusive
case class Interval(low:Int, hi:Int) {
  assert(low >= 0 && hi >= low && hi <= Intervals.maxValue)
  override def toString = "[" + low + ".." + hi + "]"
}


// represents the possible values of a variable as a list of
// intervals. by construction this list contains non-overlapping
// intervals in ascending order.
case class Intervals(ns:List[Interval]) {
  // are there no possible values?
  def isEmpty:Boolean =
    ??? // FILL ME IN

  // apply function f to every possible value
  def foreach( f:Int => Unit ): Unit =
    ??? // FILL ME IN

  // compute the intersection of these possible values and another set
  // of possible values
  def intersect( i:Intervals ): Intervals =
    ??? // FILL ME IN

  // is there only a single possible value?
  def singleton:Boolean =
    ??? // FILL ME IN

  // retrieve the single possible value (only valid if there is only a
  // single possible value to retrieve)
  def getSingleton:Int =
    ??? // FILL ME IN

  // remove a possible value from the current set of possible values
  def -( n:Int ): Intervals =
    ??? // FILL ME IN

  // return the lowest currently possible value
  def lowest:Int =
    ??? // FILL ME IN

  // return the highest currently possible value
  def highest:Int =
    ??? // FILL ME IN

  // returns all currently possible values < n
  def LT( n:Int ): Intervals =
    ??? // FILL ME IN

  // returns all currently possible values ≤ n
  def LE( n:Int ): Intervals =
    ??? // FILL ME IN

  // returns all currently possible values > n
  def GT( n:Int ): Intervals =
    ??? // FILL ME IN

  // returns all currently possible values ≥ n
  def GE( n:Int ): Intervals =
    ??? // FILL ME IN

  override def toString =
    ns.mkString(", ")
}


// companion object with useful functions
object Intervals {
  // maximum possible value
  val maxValue = 500

  // default Intervals value, i.e., the default set of possible values
  // for any variable
  val default = Intervals(List(Interval(0, maxValue)))

  // create an Intervals with a single possible value
  def apply( n:Int ): Intervals =
    ??? // FILL ME IN

  // create an Intervals with a set of contiguous possible values in
  // the range low to hi
  def apply( low:Int, hi:Int ): Intervals =
    ??? // FILL ME IN

  // Create an Intervals containing all possible values that have the
  // given relation op with the given integer n.  `max` below refers
  // to the `maxValue` constant defined above.
  //
  // examples: getIntervals(<,  10) = [[0..9]]
  //           getIntervals(=,  10) = [[10..10]]
  //           getIntervals(≠,  10) = [[0..9], [11..max]]
  //           getIntervals(<,   0) = []          // less than minimum
  //           getIntervals(>, max) = []          // greater than maximum
  //           getIntervals(=,  -1) = []          // out of range
  //           getIntervals(>,  -5) = [[0..max]]  // the whole valid range
  def getIntervals( op:ROP, n:Int ): Intervals =
    ??? // FILL ME IN
}


// constraints between variables
case class Constraint(op:ROP, x1:Variable, x2:Variable)


// a map from variables to their possible values
case class Values( vs:Map[Variable, Intervals] = Map() ) {
  // retrieve the possible values for the given variable; if the
  // variable isn't in vs then return the default set of possible
  // values
  def apply( x:Variable ): Intervals =
    ??? // FILL ME IN

  // return a mutable copy of this map
  def mutableCopy:MValues =
    ??? // FILL ME IN
}


// a mutable map from variables to their possible values
case class MValues( vs:MMap[Variable, Intervals] = MMap() ) {
  // retrieve the possible values for the given variable; if the
  // variable isn't in vs then return the default set of possible
  // values
  def apply( x:Variable ): Intervals =
    ??? // FILL ME IN

  // update the possible values of the given variable
  def update( x:Variable, i:Intervals ) {
    ??? // FILL ME IN
  }

  // clear the map
  def clear() {
    ??? // FILL ME IN
  }

  // is the map empty?
  def isEmpty:Boolean =
    ??? // FILL ME IN

  // return an immutable copy of this map
  def immutableCopy:Values =
    ??? // FILL ME IN
}


// a map from variables to the constraints mentioning that variable
case class Constraints( cs:Map[Variable, Set[Constraint]] = Map() ) {
  // retrieve the constraint set for the given variable; if the
  // variable isn't in cs then return the empty set
  def apply( x:Variable ): Set[Constraint] =
    ??? // FILL ME IN

  // add constraint c to the constraints sets for variables x1 and x2
  // mentioned in constraint c
  def +( c:Constraint ): Constraints = 
    ??? // FILL ME IN
}


// this is the actual solver, including an API used by the
// miniprolog-fd interpreter to insert constraints and query for
// satisfiable values.
case class ConstraintStore(
  values:Values = Values(),
  constraints:Constraints = Constraints()
) {

  ////////// API FOR INTERPRETER //////////

  // insert a new constraint into the constraint store. the return
  // value is None if the new constraint is inconsistent with existing
  // constraints, otherwise it is Some(the updated constraint store).
  def insert(op:ROP, v1:Value, v2:Value): Option[ConstraintStore] =
    ??? // FILL ME IN

  // query the constraint store for satisfying values for the given
  // list of variables
  def SAT(xs:List[Variable]): List[NumValue] =
    ??? // FILL ME IN


  ////////// INTERNAL METHODS //////////


  // given the current possible values for all the variables, compute
  // satisfying values for the given list of variables
  def solve( currValues:Values, xs:List[Variable] ): Option[Values] =
    ??? // FILL ME IN

  // constraint propagation: restrict the possible values of x to only
  // include values in ranges. if that makes x unsatisfiable then
  // return; if that changes the possible values of x then recursively
  // propagate those changes to the other variables.
  //
  // IMPORTANT NOTE: because mutableCV returns a default set of
  // possible values for any variable not contained in the mapping, we
  // can't just iterate through all the constraints as done in the
  // pseudocode (it would incorrectly propagate default values if
  // mutableCV is made empty in some recursive call to
  // propagate). instead, we need to explicitly check after each
  // recursive call to propagate whether mutableCV is empty, and if so
  // then explicitly return from the function.
  def propagate( mutableCV:MValues, x:Variable, ranges:Intervals) {
    ??? // FILL ME IN
  }

  // narrow the current possible values of x1 (given as rangesX1)
  // according to the constraint c, which is between x1 and some
  // variable x2 (where rangesX2 gives the possible values of x2)
  def narrow( x1:Variable, rangesX1:Intervals, c:Constraint, rangesX2:Intervals ): Intervals =
    ??? // FILL ME IN

  // reverse the meaning of a relational operator (e.g., < becomes >
  // and ≤ becomes ≥)
  def reverseOp( op:ROP ): ROP =
    op match {
      case EQ => EQ
      case NE => NE
      case LT => GT
      case LE => GE
      case GT => LT
      case GE => LE
    }
}
